{Graficky unit}
{R. Novotny}

{$DEFINE WIN32}

unit grafika;

interface

{$IFDEF WIN32}
uses graphics, math;

{$ELSE}
uses dos;
{$ENDIF}

type TPoint = record
  x:real;
  y:real;
  z:real;
end;

type TPointH = record
  x:real;
  y:real;
  z:real;
  h:real;
end;

type TMatrix = array[1..4, 1..4] of real;
type TPointMatrix = array[1..4] of real;
//type TShape = array[1..30] of TPointH;
type TVector = array[1..3] of real;

type TPoints = record
  points: array[1..50] of TPointMatrix;
  pointcount: integer;
end;  

type TEdge = array[1..2] of integer; //edge has 2 vertices

type TFace = record
  edges: array[1..50] of TEdge;  //face has at most 50 vertices
  ecount: integer; //number of edges in each face
  nvector: TVector; //normal vector
  nvector2: TVector; //normal vector 
  color: integer;
end;

type TShape = record
  faces: array[1..50] of TFace; //shapes is made of at most 50 faces
  facecount: integer;
end;

type TEdges = record
  edges: array[1..50] of TEdge; //shapes is made of at most 50 faces
  edgecount: integer;
end;

//upratat
var edges: TEdges;
    points: TPoints;

//semienko v seedfill-e
type TFillSeed=record
  x:integer;
  y:integer;
end;

//stack v seedfille
var seedstack:record
  items: array[1..255] of TFillSeed;
  size: integer;
end;

const xlatematrix:TMatrix = ((1,0,0,0),(0,1,0,0),(0,0,1,0), (MAXINT, MAXINT, MAXINT, 1));
const scalematrix:TMatrix = ((MAXINT,0,0,0),(0,MAXINT,0,0),(0,0,MAXINT,0),(0, 0, 0, 1));
const rotatematrixX:TMatrix = ((1, 0, 0, 0), (0, MAXINT, MAXINT, 0), (0, MAXINT, MAXINT, 0), (0, 0, 0, 1));
// to be used in 2D
const rotatematrixY:TMatrix = ((MAXINT,0,MAXINT,0), (0, 1, 0, 0), (MAXINT,0,MAXINT,0), (0, 0, 0, 1));
const rotatematrixZ:TMatrix = ((MAXINT,MAXINT,0,0), (MAXINT,MAXINT,0,0), (0, 0, 1, 0), (0, 0, 0, 1));

var bodp:TPointMatrix = (0, 0, -50000, 1);
//namiesto vyplnania len zobrazovat pociatocne body fillu
var debugFill:boolean = false;
//ak true, bude pouzity GDI floodfill, inak vlastny SeedFill()
var useGDIFill:boolean = false;

{$IFDEF WIN32}
var mem:TCanvas;
{$ELSE}
var mem:array [0..$7FFF] of byte absolute $A000:0000;
{$ENDIF}

{$IFDEF WIN32}
procedure init(canvas:TCanvas);
{$ELSE}
procedure init(mode:byte);
{$ENDIF}

{$IFDEF WIN32}
procedure usecka_b(x1, y1, x2, y2: integer; farba: TColor);
{$ELSE}
procedure usecka_b(x1, y1, x2, y2: integer; farba: byte);
{$ENDIF}

{$IFDEF WIN32}
procedure bod_pc(x,y:integer; farba:TColor);
{$ELSE}
procedure bod_pc(x,y:integer; farba:byte);
{$ENDIF}

{$IFDEF WIN32}
function farbabodu_pc(x,y:integer):TColor;
{$ELSE}
function farbabodu_pc(x,y:integer):byte;
{$ENDIF}

procedure PointToPointH(pt:TPoint; var pth:TPointH);
procedure PointHToPoint(pth:TPointH; var pt:TPoint);
procedure PointMatrixTimesMatrix(pt:TPointMatrix; m:TMatrix; var result:TPointMatrix);
procedure PointHToPointMatrix(pth:TPointH; var ptm:TPointMatrix);
procedure PointMatrixToPointH(ptm:TPointMatrix; var pth:TPointH);
procedure multiplyMatrices(m,n: TMatrix; var p: TMatrix);
procedure getTranslationMatrix(dx, dy, dz:integer; var m:TMatrix);
procedure getRotationMatrixX(angle:real; var m:TMatrix);
procedure getRotationMatrixY(angle:real; var m:TMatrix);
procedure getRotationMatrixZ(angle:real; var m:TMatrix);
procedure translatePointH(pt:TPointH; dx, dy, dz:integer; var xpth:TPointH);
procedure scalePointH(pt:TPointH; xscale, yscale, zscale:real; var xpth:TPointH);
procedure rotatePointHX(pt:TPointH; angle:real; var xpth:TPointH);
procedure rotatePointHY(pt:TPointH; angle:real; var xpth:TPointH);
procedure rotatePointHZ(pt:TPointH; angle:real; var xpth:TPointH);

procedure rotatePointHExX(pt:TPointH; angle:real; xoffset, yoffset, zoffset:integer; var xpth:TPointH);
procedure rotatePointHExY(pt:TPointH; angle:real; xoffset, yoffset, zoffset:integer; var xpth:TPointH);
procedure rotatePointHExZ(pt:TPointH; angle:real; xoffset, yoffset, zoffset:integer; var xpth:TPointH);

function isVisibleFace(face: TFace; bodv: TPointMatrix; var angle: real): boolean;
procedure getVector(pt1, pt2: TPointMatrix; var v: TVector);
procedure vectorToPointH(v: TVector; var pth: TPointH);
procedure PointHToVector(pth: TPointH; var v:TVector);
procedure vectorCrossProduct(v1, v2: TVector; var v: TVector);

procedure fillFace(face: TFace; Canvas: TCanvas; color:TColor);
procedure SeedFill(x, y:integer; color: TColor);

implementation

{$IFDEF WIN32}
procedure init(canvas:TCanvas);
begin
  mem:=canvas;
end;
{$ELSE}
procedure init(mode:byte);
begin
     asm
     mov ah,0;
     mov al,mode;
     int $10;
     end;
end;
{$ENDIF}

{$IFDEF WIN32}
function farbabodu_pc(x,y:integer):TColor;
begin
  farbabodu_pc:=mem.Pixels[x,y]
end;

{$ELSE}

function farbabodu_pc(x,y:integer):byte;
begin
  farbabodu_pc:=mem[320*y+x];
end;
{$ENDIF}

{$IFDEF WIN32}
procedure usecka_b(x1, y1, x2, y2: integer; farba: TColor);
{$ELSE}
procedure usecka_b(x1, y1, x2, y2: integer; farba: byte);
{$ENDIF}

var i,dx,dy,pixelov,d,dp1,dp2,x,posunx1,posunx2,y,posuny1,posuny2: integer;
begin
  dx:=abs(x2-x1);
  dy:=abs(y2-y1);
  {uhol <= 45}
  if dx>=dy then begin
    pixelov:=dx+1;
    d:=(2*dy)-dx;
    dp1:=dy*2;
    dp2:=(dy-dx)*2;
    posunx1:=1;
    posunx2:=1;
    posuny1:=0;
    posuny2:=1;
  end
  {uhol > 45}
  else begin
    pixelov:=dy+1;
    d:=(2*dx)-dy;
    dp1:=dx*2;
    dp2:=(dx-dy)*2;
    posunx1:=0;
    posunx2:=1;
    posuny1:=1;
    posuny2:=1;
  end;
  {druhy bod je viac vpravo ako prvy bod, otacame suradnice}
  if x1 > x2 then begin
    posunx1:= -posunx1;
    posunx2:= -posunx2;
  end;
  {prvy bod je vyssie ako druhy bod, otacame suradnice}
  if y1 > y2 then begin
    posuny1:= -posuny1;
    posuny2:= -posuny2;
  end;
  {kreslime}
  x:=x1;
  y:=y1;
  for i:=1 to pixelov do begin
    bod_pc(x, y, farba);
    if d < 0 then begin
      d:=d+dp1;
      x:=x+posunx1;
      y:=y+posuny1;
    end
    else begin
      d:=d+dp2;
      x:=x+posunx2;
      y:=y+posuny2;
    end;
  end;
end;

{$IFDEF WIN32}
procedure bod_pc(x,y:integer; farba:TColor);
begin
    mem.Pixels[x, y]:=farba;
    //mem.Ellipse(x - 2, y - 2, x + 2, y + 2);
end;
{$ELSE}
procedure bod_pc(x,y:integer; farba:byte);
Var Reg: Registers;
begin
  Reg.AH := $0C; {service no}
  Reg.AL := farba; {pt color}
  Reg.BH := $0; {display page no (ignored in 12h}
  Reg.CX := x; {x-coord}
  Reg.DX := y; {y-coord}
  Intr($10, Reg); {call interrupt}
end;
{$ENDIF}

{-----------------------------------------------------------}

procedure PointToPointH(pt:TPoint; var pth:TPointH);
begin
  pth.x:=pt.x;
  pth.y:=pt.y;
  pth.z:=pt.z;
  pth.h:=1;
end;

procedure PointHToPoint(pth:TPointH; var pt:TPoint);
  var xx,yy,zz,hh:real;
begin
  xx:=pth.x;
  yy:=pth.y;
  zz:=pth.z;
  hh:=pth.h;

  pt.x:=xx / hh;
  pt.y:=yy / hh;
  pt.z:=zz / hh;
end;

procedure PointMatrixTimesMatrix(pt:TPointMatrix; m:TMatrix; var result:TPointMatrix);
  var i,j,k:integer;
begin
  for i:=1 to 4 do begin
    result[i]:=0;
    for j:=1 to 4 do begin
      result[i]:=result[i] + (pt[j]*m[j,i]);
    end;
  end;
end;

procedure PointHToPointMatrix(pth:TPointH; var ptm:TPointMatrix);
begin
  ptm[1]:=pth.x;
  ptm[2]:=pth.y;
  ptm[3]:=pth.z;
  ptm[4]:=pth.h;
end;

procedure PointMatrixToPointH(ptm:TPointMatrix; var pth:TPointH);
begin
  pth.x := ptm[1];
  pth.y := ptm[2];
  pth.z := ptm[3];
  pth.h := ptm[4];
end;

procedure multiplyMatrices(m,n: TMatrix; var p: TMatrix);
  var i,j,k:byte;
begin
  for i:=1 to 4 do begin
    for j:=1 to 4 do begin
      p[i,j]:=0;
      for k:=1 to 4 do begin
        p[i,j]:=p[i,j]+(m[i,k]*n[k,j]);
     end;
    end;
  end;
end;

procedure getTranslationMatrix(dx, dy, dz:integer; var m:TMatrix);
begin
  m := xlateMatrix;
  m[4,1] := dx;
  m[4,2] := dy;
  m[4,3] := dz;
end;

procedure getRotationMatrixX(angle:real; var m:TMatrix);
  var radangle:real;
      cosa, sina:real;
begin
  radangle:=angle * (pi/180);
  cosa:=cos(radangle);
  sina:=sin(radangle);
  m := rotateMatrixX;
  m[2,2] := cosa;  m[2,3] := sina;
  m[3,2] := -sina; m[3,3] := cosa;
end;

procedure getRotationMatrixY(angle:real; var m:TMatrix);
  var radangle:real;
      cosa, sina:real;
begin
  radangle:=angle * (pi/180);
  cosa:=cos(radangle);
  sina:=sin(radangle);
  m := rotateMatrixY;
  m[1,1] := cosa;  m[1,3] := sina;
  m[3,1] := -sina; m[3,3] := cosa;
end;

procedure getRotationMatrixZ(angle:real; var m:TMatrix);
  var radangle:real;
      cosa, sina:real;
begin
  radangle:=angle * (pi/180);
  cosa:=cos(radangle);
  sina:=sin(radangle);
  m := rotateMatrixZ;
  m[1,1] := cosa;  m[1,2] := sina;
  m[2,1] := -sina; m[2,2] := cosa;
end;


procedure translatePointH(pt:TPointH; dx,dy,dz:integer; var xpth:TPointH);
  var ptm:TPointMatrix;
      resultPtMatrix:TPointMatrix;
      pth:TPointH;
      xlationMatrix:TMatrix;
begin
  getTranslationMatrix(dx, dy, dz, xlationMatrix);
  PointHToPointMatrix(pt, ptm);
  PointMatrixTimesMatrix(ptm, xlationmatrix, resultPtMatrix);
  PointMatrixToPointH(resultPtMatrix, xpth);
end;

procedure scalePointH(pt:TPointH; xscale, yscale, zscale:real; var xpth:TPointH);
  var ptm:TPointMatrix;
      resultPtMatrix:TPointMatrix;
      pth:TPointH;
      skaleMatrix:TMatrix;
begin
  skaleMatrix := scaleMatrix;
  skaleMatrix[1,1] := xscale;
  skaleMatrix[2,2] := yscale;
  skaleMatrix[3,3] := zscale;  
  PointHToPointMatrix(pt, ptm);
  PointMatrixTimesMatrix(ptm, skaleMatrix, resultPtMatrix);
  PointMatrixToPointH(resultPtMatrix, xpth);
end;

procedure rotatePointHX(pt:TPointH; angle:real; var xpth:TPointH);
  var ptm:TPointMatrix;
      resultPtMatrix:TPointMatrix;
      pth:TPointH;
      rotMatrix:TMatrix;
      cosa:real;
      sina:real;
      radangle:real;
begin
  radangle:=angle * (pi/180);
  cosa:=cos(radangle);
  sina:=sin(radangle);
  getRotationMatrixX(angle, rotMatrix);
  PointHToPointMatrix(pt, ptm);
  PointMatrixTimesMatrix(ptm, rotMatrix, resultPtMatrix);
  PointMatrixToPointH(resultPtMatrix, xpth);
end;

procedure rotatePointHY(pt:TPointH; angle:real; var xpth:TPointH);
  var ptm:TPointMatrix;
      resultPtMatrix:TPointMatrix;
      pth:TPointH;
      rotMatrix:TMatrix;
      cosa:real;
      sina:real;
      radangle:real;
begin
  radangle:=angle * (pi/180);
  cosa:=cos(radangle);
  sina:=sin(radangle);
  getRotationMatrixY(angle, rotMatrix);
  PointHToPointMatrix(pt, ptm);
  PointMatrixTimesMatrix(ptm, rotMatrix, resultPtMatrix);
  PointMatrixToPointH(resultPtMatrix, xpth);
end;

procedure rotatePointHZ(pt:TPointH; angle:real; var xpth:TPointH);
  var ptm:TPointMatrix;
      resultPtMatrix:TPointMatrix;
      pth:TPointH;
      rotMatrix:TMatrix;
      cosa:real;
      sina:real;
      radangle:real;
begin
  radangle:=angle * (pi/180);
  cosa:=cos(radangle);
  sina:=sin(radangle);
  getRotationMatrixZ(angle, rotMatrix);
  PointHToPointMatrix(pt, ptm);
  PointMatrixTimesMatrix(ptm, rotMatrix, resultPtMatrix);
  PointMatrixToPointH(resultPtMatrix, xpth);
end;

procedure rotatePointHExX(pt:TPointH; angle:real; xoffset, yoffset, zoffset:integer; var xpth:TPointH);
  var ptm:TPointMatrix;
      resultPtMatrix:TPointMatrix;
      pth:TPointH;
      rotMatrix:TMatrix;
      cosa:real;
      sina:real;
      radangle:real;
      mx1, mx2, mr, mt1: TMatrix;
begin
  getTranslationMatrix(-xoffset, -yoffset, -zoffset, mx1);
  getRotationMatrixX(angle, mr);
  getTranslationMatrix(xoffset, yoffset, zoffset, mx2);

  multiplyMatrices(mx1, mr, mt1);
  multiplyMatrices(mt1, mx2, rotMatrix);

  PointHToPointMatrix(pt, ptm);
  PointMatrixTimesMatrix(ptm, rotMatrix, resultPtMatrix);
  PointMatrixToPointH(resultPtMatrix, xpth);             
end;

procedure rotatePointHExY(pt:TPointH; angle:real; xoffset, yoffset, zoffset:integer; var xpth:TPointH);
  var ptm:TPointMatrix;
      resultPtMatrix:TPointMatrix;
      pth:TPointH;
      rotMatrix:TMatrix;
      cosa:real;
      sina:real;
      radangle:real;
      mx1, mx2, mr, mt1: TMatrix;
begin
  getTranslationMatrix(-xoffset, -yoffset, -zoffset, mx1);
  getRotationMatrixY(angle, mr);
  getTranslationMatrix(xoffset, yoffset, zoffset, mx2);

  multiplyMatrices(mx1, mr, mt1);
  multiplyMatrices(mt1, mx2, rotMatrix);

  PointHToPointMatrix(pt, ptm);
  PointMatrixTimesMatrix(ptm, rotMatrix, resultPtMatrix);
  PointMatrixToPointH(resultPtMatrix, xpth);
end;

procedure rotatePointHExZ(pt:TPointH; angle:real; xoffset, yoffset, zoffset:integer; var xpth:TPointH);
  var ptm:TPointMatrix;
      resultPtMatrix:TPointMatrix;
      pth:TPointH;
      rotMatrix:TMatrix;
      cosa:real;
      sina:real;
      radangle:real;
      mx1, mx2, mr, mt1: TMatrix;
begin
  getTranslationMatrix(-xoffset, -yoffset, -zoffset, mx1);
  getRotationMatrixZ(angle, mr);
  getTranslationMatrix(xoffset, yoffset, zoffset, mx2);

  multiplyMatrices(mx1, mr, mt1);
  multiplyMatrices(mt1, mx2, rotMatrix);

  PointHToPointMatrix(pt, ptm);
  PointMatrixTimesMatrix(ptm, rotMatrix, resultPtMatrix);
  PointMatrixToPointH(resultPtMatrix, xpth);
end;

procedure vectorCrossProduct(v1, v2: TVector; var v: TVector);
begin
  v[1] := -((v1[2] * v2[3]) - (v1[3] * v2[2]));
  v[2] := (v1[3] * v2[1]) - (v1[1] * v2[3]);
  v[3] := (v1[1] * v2[2]) - (v1[2] * v2[1]);
end;

function vectorDotProduct(v1, v2: TVector):real;
begin
  vectorDotProduct := v1[1] * v2[1] + v1[2] * v2[2] + v1[3] * v2[3];
end;


procedure getVector(pt1, pt2: TPointMatrix; var v: TVector);
begin
  v[1] := pt2[1] - pt1[1];
  v[2] := pt2[2] - pt1[2];
  v[3] := pt2[3] - pt1[3];
end;

procedure vectorToPointH(v: TVector; var pth: TPointH);
  var i:integer;
begin
  pth.x := v[1];
  pth.y := v[2];
  pth.z := v[3];
  pth.h:=1;
end;

procedure PointHToVector(pth: TPointH; var v:TVector);
  var xx,yy,zz,hh:real;
begin
  xx:=pth.x;
  yy:=pth.y;
  zz:=pth.z;
  hh:=pth.h;

  v[1]:=xx / hh;
  v[2]:=yy / hh;
  v[3]:=zz / hh;
end;

function vectorNorm(v: TVector):real;
begin
  vectorNorm:=sqrt(v[1]*v[1] + v[2]*v[2] + v[3]*v[3]);
end;

function isVisibleFace(face: TFace; bodv: TPointMatrix; var angle: real): boolean;
  var v,w,u: TVector;
      x,y,z, d:real;
      vp, vs: real;
      cosangle: real;
      degangle: real;
      pth: TPointH;
begin
  //opravit!
  u := face.nvector;
  x:=points.points[face.edges[1][1]][1];
  y:=points.points[face.edges[1][1]][2];
  z:=points.points[face.edges[1][1]][3];
  d := -(u[1]*x  + u[2]*y + u[3]*z);
  vp := u[1]*bodp[1] + u[2]*bodp[2] + u[3]*bodp[3] + d;
  vs := u[1]*bodv[1] + u[2]*bodv[2] + u[3]*bodv[3] + d;
  v[1]:=bodp[1];  v[2]:=bodp[2];   v[3]:=bodp[3];
  try
    angle := arccos(abs(vectorDotProduct(u, v) / (vectorNorm(v) * vectorNorm(u))));
  except
    angle := 0;
  end;
  if(vp * vs > 0) then begin
   isVisibleFace:=true;
  end
  else begin
    isVisibleFace:=false;
  end;

  {
  v[1] := bodp[1]; v[2] := bodp[2]; v[3] := bodp[3];
  try
    angle := arccos(vectorDotProduct(u, v) / (vectorNorm(v) * vectorNorm(u)));
  except
    isVisibleFace:=false;
    exit;
  end;
  degangle:= angle * (180/pi);
  if (-90<degangle) and (degangle<90) then begin
    isVisibleFace:=true;
  end
  else begin
    isVisibleFace:=false;
  end;
  }
end;

procedure floodFill(x,y:integer; farba:TColor; farbahranice:TColor);
begin
  if(x>=0) and (y>=0) and (x < mem.ClipRect.Right) and ( y < mem.ClipRect.Bottom) then begin
    if (farbabodu_pc(x,y)<>farbahranice) and (farbabodu_pc(x,y)<>farba) then begin
      bod_pc(x,y,farba);
      floodFill(x-1, y, farba, farbahranice);
      floodFill(x, y-1, farba, farbahranice);
      floodFill(x, y+1, farba, farbahranice);
      floodFill(x+1, y, farba, farbahranice);
    end;
  end;
end; {maluj}

function reorientEdge(edge:TEdge):TEdge;
begin
  Result[1]:=edge[2];
  Result[2]:=edge[1];
end;

function isInRectangle(pt:TPointMatrix; left, top, right, bottom:integer):boolean;
begin
  Result:=false;
  if (pt[1] >= left) and (pt[1] <= right) and
     (pt[2] >= top) and (pt[2] <= bottom) then
     Result:=true;
end;

procedure fillFace(face: TFace; Canvas: TCanvas; color:TColor);
var v, w, u:TVector;
    pt: TPointMatrix;
    c: TColor;
    startptindex:integer;
    i,j:integer;
    k,ii:integer;
    secondedge:TEdge;
    tempedge:TEdge;
    firstedgeindex:integer;
    firstedge:TEdge;
    usize:real;
    firstptfound:boolean;
    hasvisiblept:boolean;
begin
  firstptfound:=false;
  hasvisiblept:=false;
  for i:=1 to face.ecount do begin
    for j:=1 to 2 do begin
      startptindex:=face.edges[i][j];
      if isInRectangle(points.points[startptindex], 0 - 150, 0 - 150, mem.ClipRect.Right - 1 - 150, mem.ClipRect.Bottom - 1 - 150) then begin
        hasvisiblept:=true;
        for k:=1 to face.ecount do begin
          if (face.edges[k][1] = startptindex) then begin
            firstedge:=face.edges[k];
            firstedgeindex:=k;
            firstptfound:=true;
            break;
          end;
          if (face.edges[k][2] = startptindex) then begin
            firstedge:=ReorientEdge(face.edges[k]);
            firstedgeindex:=k;
            firstptfound:=true;
            break;
          end;
        end;
        break;
      end;
    end;
    if firstptfound then
      break;
  end;
  if not hasvisiblept then exit;

  for i:=1 to face.ecount do begin
    if i<>firstedgeindex then begin
      if face.edges[i][1] = startptindex then begin //hrana zacina v pociatocnom bode
        secondEdge := face.edges[i];
        break;
      end
      else begin
        if face.edges[i][2] = startptindex then begin //hrana konci v pociatocnom bode
          secondEdge[1] := face.edges[i][2];
          secondEdge[2] := face.edges[i][1];
          break;
        end;
      end;
    end;
  end;

  getVector(points.points[firstEdge[1]], points.points[firstEdge[2]], v);
  getVector(points.points[secondEdge[1]],points.points[secondEdge[2]], w);
  u[1] := v[1] + w[1];
  u[2] := v[2] + w[2];
  u[3] := v[3] + w[3];
  usize:=vectorNorm(u);

  // \vec{u} =\vec{v} + \vec{w}
  k:=2;
  repeat
    pt[1] := points.points[firstEdge[1]][1] + (u[1] / k);
    pt[2] := points.points[firstEdge[1]][2] + (u[2] / k);
    pt[3] := points.points[firstEdge[1]][3] + (u[3] / k);
    k:=k*2;
  until (farbabodu_pc(150 + Round(pt[1]), 150 + Round(pt[2])) = clWhite) or
        (farbabodu_pc(150 + Round(pt[1]), 150 + Round(pt[2])) = 1); 

  //Canvas.Pixels[Round(pt[1]), Round(pt[2])] := clRed;
  if debugFill then begin
    Canvas.Ellipse(150 + Round(pt[1]) - 4, 150 + Round(pt[2]) - 4, 150 + Round(pt[1]) + 4, 150 + Round(pt[2]) + 4);
  end;
  if useGDIFill then begin
    c:=canvas.brush.Color;
    canvas.brush.Color := color;
    canvas.FloodFill(150 + Round(pt[1]), 150 + Round(pt[2]), clWhite, fsSurface);
    canvas.brush.Color :=c;
  end
  else begin
    SeedFill(150 + Round(pt[1]), 150 + Round(pt[2]), color);
  end;


  {
  floodFill(150 + Round(pt[1]), 150 + Round(pt[2]), color, clBlack);
  }

end;

procedure SeedFill(x, y:integer; color: TColor);
//inspirovane navodom v elektronickych skriptach
//Algoritmy pocitacove grafiky (ALP),
//J. Sloup a kol., CVUT 1999/200
  var currseed: TFillSeed;
      i:integer;
      x_r,x_l:integer;
begin
  //1. Vlo uivatelem zadan semnko do zsobnku
  inc(seedstack.size);
  seedstack.items[seedstack.size].x:=x;
  seedstack.items[seedstack.size].y:=y;
  //2. Dokud nen zsobnk przdn
  while seedstack.size > 0 do begin
    // 2.a) vyjmi semnko (se souadnicemi [x,y]) ze zsobnku
    currseed:=seedstack.items[seedstack.size];
    dec(seedstack.size);
    // 2.b) nalezni hranici xL a xR na dku y ve nejblim okol bodu [x,y]
    for i:=currseed.x to mem.ClipRect.Right do begin
      if farbabodu_pc(i, currseed.y) <> clWhite then begin
        x_R:=i-1;
        break;
      end;
    end;
    // hranica xR
    for i:=currseed.x downto 1 do begin
      if farbabodu_pc(i, currseed.y) <> clWhite then begin
        x_L:=i+1;
        break;
      end;
    end;
    // 2.c) nakresli seku [xL,y] - [xR,y]
    usecka_b(x_L, currseed.y, x_r, currseed.y, color);
    // 2.d) na (vy) sece [xL,y + 1] - [xR,y + 1] hledej souvisl vnitn seky
    // a pro kad z nich vlo do zsobnku souadnice jednoho vnitnho bodu
    i:=x_L;
    repeat
      if farbabodu_pc(i, currseed.y + 1) = clWhite then begin
        while(farbabodu_pc(i, currseed.y + 1) = clWhite) do begin
          inc(i);
        end;
        //i-ty bod je okraj, ale (i-1)-vy je stale nevyfarbeny
        //vloz dalsie semienko do stacku
        inc(seedstack.size);
        seedstack.items[seedstack.size].x:=i-1;
        seedstack.items[seedstack.size].y:=currseed.y+1;
      end;
      inc(i);
    until i>=x_R;
    // 2.e) na (ni) sece [xL,y - 1] - [xR,y - 1] hledej souvisl vnitn seky
    //a pro kad z nich vlo do zsobnku souadnice jednoho vnitnho bodu
    i:=x_L;
    repeat
      if farbabodu_pc(i, currseed.y - 1) = clWhite then begin
        while(farbabodu_pc(i, currseed.y - 1) = clWhite) do begin
          inc(i);
        end;
        //i-ty bod je okraj, ale (i-1)-vy je stale nevyfarbeny
        //vloz dalsie semienko do stacku
        inc(seedstack.size);
        seedstack.items[seedstack.size].x:=i-1;
        seedstack.items[seedstack.size].y:=currseed.y-1;
      end;
      inc(i);
    until i>=x_R;
  end;
end;

end.
